////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#ifndef CRNLP_GENERAL_SLLS_H 
#define CRNLP_GENERAL_SLLS_H

#include <stdbool.h>
#include <stdint.h>
#include "nesting.h"
#include "cycles.h"
#include "queue.h"
#include "setup.h"

// Reader-Reader logic
#define ABITS 0x00FF // lock server 0 phase bits in S
#define BBITS 0xFF00 // lock server 1 phase bits in S

#define APHID 0x0001 // lock server 0 phase ID bit in S
#define BPHID 0x0100 // lock server 1 phase ID bit in S

#define APRES 0x0002 // lock server 0 present bit in S
#define BPRES 0x0200 // lock server 1 present bit in S

typedef struct {
    resource_mask_t       requested;
    uint32_t              length;
    uint32_t              request_id;
    uint_fast8_t volatile needs_service;
    uint_fast8_t          done;
    uint_fast8_t volatile enabled; // set by the server to indicate satisfaction
    uint8_t               _buf[45]; // separate nodes per cache line
} crnlp_general_slls_request_t;

// A request Ri will correspond to |Di| nodes
typedef struct crnlp_general_slls_node_t {
    // Time the request will be satisfied
    uint64_t start_time;

    // Double-LL pointers
    volatile struct crnlp_general_slls_node_t * volatile prev;
    volatile struct crnlp_general_slls_node_t * volatile next;

    // Request information
    crnlp_general_slls_request_t *request;

} crnlp_general_slls_node_t;

typedef struct {
    // Head and tail pointers
    volatile crnlp_general_slls_node_t * volatile head;
    volatile crnlp_general_slls_node_t * volatile tail;
} crnlp_general_slls_queue_t;

typedef struct
{
    // Set of requests
    __attribute__((aligned(64))) crnlp_general_slls_request_t requests[QUEUE_LENGTH];

    // Pool of available nodes, MAX_REQUESTED per request
    crnlp_general_slls_node_t nodes[QUEUE_LENGTH][MAX_REQUESTED];

    // Set of resource queues, one per resource
    volatile crnlp_general_slls_queue_t queues[RES_COUNT];

    // Count of total pending and satisfied requests
    uint_fast8_t pending_requests;
    uint_fast8_t satisfied_requests;

    // Built-in Reader-Reader logic to arbitrate between servers
    volatile uint32_t rr_wave; // just all 0s or all 1s

    // Whether this server is currently enabled
    volatile uint8_t enabled;

} crnlp_general_slls_server_lock_t;

typedef struct
{
    // One lock server per socket
    crnlp_general_slls_server_lock_t servers[NUM_SERVERS];

    // Built-in Reader-Reader logic to arbitrate between servers
    __attribute__((aligned(64))) volatile unsigned int S;
    unsigned int _buf[15];

} crnlp_general_slls_lock_t;

// Res-queue + global LS C-RNLP initialization, lock, and unlock
void crnlp_general_slls_init(crnlp_general_slls_lock_t* lock);

uint64_t *crnlp_general_slls_lock(crnlp_general_slls_lock_t* lock, uint8_t ind);
uint64_t *crnlp_general_slls_unlock(crnlp_general_slls_lock_t* lock, uint8_t ind);

#endif // CRNLP_GENERAL_SLLS_H
