////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include <sched.h>
#include <inttypes.h>
#ifdef __APPLE__
#include <sys/uio.h>
#else
#include <sys/io.h>
#endif
#include <sys/mman.h>

#include "nesting.h"
#include "queue.h"
#include "rnlp.h"
#include "crnlp_uniform.h"
#include "crnlp_uniform_fgls.h"
#include "crnlp_uniform_flls.h"
#include "crnlp_general.h"
#include "crnlp_general_fgls.h"
#include "crnlp_general_flls.h"
#include "r2lp.h"
#include "r3lp.h"
#include "r4lp.h"
#include "r5lp.h"
#include "r6lp.h"
#include "r7lp.h"
#include "r8lp.h"
#include "r9lp.h"
#include "r10lp.h"
#include "ticket.h"
#include "mem.h"
#include "setup.h"

#define GHZ PROC_SPEED

int nr_cpus;
int nr_requested;
int nr_resources;
int cs_length;
int cs_alpha;
int cs_beta;

/*
    Algorithms:
      1: MCS
      2: RNLP
      3: U-C-RNLP
      6: U-C-RNLP with Floating Global Lock Server
      7: U-C-RNLP with Floating Local Lock Servers
      8: G-C-RNLP
     11: G-C-RNLP with Floating Global Lock Server
     12: G-C-RNLP with Floating Local Lock Servers
     ------------
     13: R3LP
     14: R4LP
     15: R5LP
     16: R6LP
     17: R7LP
     18: R8LP
     19: R9LP
     20: R10LP
     21: R2LP
*/
int alg;

mcs_lock_t                mcs;
rnlp_lock_t               rnlp;
crnlp_uniform_lock_t      crnlp_uniform;
crnlp_uniform_fgls_lock_t crnlp_uniform_fgls;
crnlp_uniform_flls_lock_t crnlp_uniform_flls;
crnlp_general_lock_t      crnlp_general;
crnlp_general_fgls_lock_t crnlp_general_fgls;
crnlp_general_flls_lock_t crnlp_general_flls;
r2lp_t                    r2lp;
r3lp_t                    r3lp;
r4lp_t                    r4lp;
r5lp_t                    r5lp;
r6lp_t                    r6lp;
r7lp_t                    r7lp;
r8lp_t                    r8lp;
r9lp_t                    r9lp;
r10lp_t                   r10lp;

ticket_lock_t csvlock;
FILE *csvfile;

int shared_objects[sizeof(resource_mask_t)*8];

static volatile int counter = 0;

struct thread_context {
    int id;
    int* objects;
};

#ifndef __APPLE__
int migrate_thread_to_cpu(int target_cpu)
{
    cpu_set_t *cpu_set;
    size_t sz;
    int ret;

    cpu_set = CPU_ALLOC(nr_cpus);
    sz = CPU_ALLOC_SIZE(nr_cpus);
    CPU_ZERO_S(sz, cpu_set);
    CPU_SET_S(target_cpu, sz, cpu_set);

    ret = sched_setaffinity(0, sz, cpu_set);

    CPU_FREE(cpu_set);

    return ret;
}
#endif

unsigned int choose_random_resource(int num_resources_total)
{
    return rand() % num_resources_total;
}

resource_mask_t random_resources(int requested, int available)
{
    int i, j, tmp;
    bool flag;
    int choices[nr_resources];
    resource_mask_t retval = 0;
    for (i = 0; i < requested; i++)
    {
        do
        {
            tmp = (rand() % available);
            flag = true;

            for (j = 0; j < i; j++)
            {
                if(tmp == choices[j])
                    flag = false;
            }
        } while (!flag);

        choices[i] = tmp;
        retval |= (1l << tmp);
    }

    return retval;
}

float random_cs_length(int lmax, int alpha, int beta)
{
    if ((rand() % 100) <= alpha)
    {
        return (beta / 100.0f) * lmax;
    }
    else
    {
        return lmax;
    }
}

void* rt_thread(void* args)
{
    crnlp_uniform_node_t *crnlp_uniform_self;
    crnlp_uniform_fgls_node_t *crnlp_uniform_fgls_self;
    crnlp_uniform_flls_node_t *crnlp_uniform_flls_self;
    crnlp_general_request_t *crnlp_general_self;
    crnlp_general_fgls_request_t *crnlp_general_fgls_self;
    crnlp_general_flls_request_t *crnlp_general_flls_self;
    rnlp_node_t rnlp_self;
    mcs_node_t mcs_self;
    int i,j,k,l,r,server_id;
    struct thread_context* ctx = (struct thread_context*) args;
    resource_mask_t tmp;
    uint64_t max_l = 0, max_u = 0;
    uint64_t sum_l = 0, sum_u = 0;
    uint64_t lock[BUFSIZE], unlock[BUFSIZE];
    uint64_t block[BUFSIZE];
    uint64_t end_time;
    resource_mask_t requests[TRIALS];
    float cs_lengths[TRIALS];
    unsigned int groups[TRIALS];
    uint64_t *ret_tmp = NULL;

#ifndef __APPLE__
    migrate_thread_to_cpu(CPU_MAP[ctx->id]);
#endif

    // Determine resources at init time so as not to alter the runtime demand
    // for the lock
    for (i = 0; i < TRIALS; i++)
    {
        requests[i] = random_resources(nr_requested, nr_resources);

        cs_lengths[i] = random_cs_length(cs_length, cs_alpha, cs_beta);

	groups[i] = choose_random_resource(nr_resources); // for RKLP
    }

    crnlp_uniform_self = &(crnlp_uniform.requests[ctx->id]);
    crnlp_uniform_fgls_self = &(crnlp_uniform_fgls.requests[ctx->id]);
    crnlp_general_self = &(crnlp_general.requests[ctx->id]);
    crnlp_general_fgls_self = &(crnlp_general_fgls.requests[ctx->id]);

    server_id = ctx->id / (QUEUE_LENGTH / NUM_SERVERS);
    crnlp_uniform_flls_self = &(crnlp_uniform_flls.servers[server_id].requests[ctx->id]);
    crnlp_general_flls_self = &(crnlp_general_flls.servers[server_id].requests[ctx->id]);

    for (l = 0; l < TRIALS/BUFSIZE; l++)
    {
        __sync_fetch_and_add(&counter, 1);
        while (counter != nr_cpus*(l+1))
        {
            cpu_relax();
        }

        mb();
    
        for (i = 0; i < BUFSIZE; i++)
        {
            k = l*BUFSIZE + i;

            switch (alg)
            {
            case 1: // MCS
                mcs_self.next = NULL;

                ret_tmp = mcs_lock(&mcs, &mcs_self);

                break;
            case 2: // RNLP
                rnlp_self.next = NULL;
	        rnlp_self.blocked = 1;
                rnlp_self.requested = requests[k];

                ret_tmp = rnlp_lock(&rnlp, &rnlp_self);

                break;
            case 3: // U-C-RNLP
                crnlp_uniform_self->requested = requests[k];
                crnlp_uniform_self->next = -1;
                crnlp_uniform_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_uniform_lock(&crnlp_uniform, ctx->id);

                break;
            case 6: // U-C-RNLP + FGLS
                crnlp_uniform_fgls_self->requested = requests[k];
                crnlp_uniform_fgls_self->next = -1;
                crnlp_uniform_fgls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_uniform_fgls_lock(&crnlp_uniform_fgls, ctx->id);

                break;
            case 7: // U-C-RNLP + FLLS
                crnlp_uniform_flls_self->requested = requests[k];
                crnlp_uniform_flls_self->next = -1;
                crnlp_uniform_flls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_uniform_flls_lock(&crnlp_uniform_flls, ctx->id);

                break;
            case 8: // G-C-RNLP
                crnlp_general_self->requested = requests[k];
                crnlp_general_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_general_lock(&crnlp_general, ctx->id);

                break;
            case 11: // G-C-RNLP + FGLS
                crnlp_general_fgls_self->requested = requests[k];
                crnlp_general_fgls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_general_fgls_lock(&crnlp_general_fgls, ctx->id);

                break;
            case 12: // G-C-RNLP + FLLS
                crnlp_general_flls_self->requested = requests[k];
                crnlp_general_flls_self->length = (int)(GHZ * cs_lengths[k]);

                ret_tmp = crnlp_general_flls_lock(&crnlp_general_flls, ctx->id);

                break;
            case 13: // R3LP
                ret_tmp = r3lp_lock(&r3lp, groups[k]);

                break;
            case 14: // R4LP
                ret_tmp = r4lp_lock(&r4lp, groups[k]);

                break;
            case 15: // R5LP
                ret_tmp = r5lp_lock(&r5lp, groups[k]);

                break;
            case 16: // R6LP
                ret_tmp = r6lp_lock(&r6lp, groups[k]);

                break;
            case 17: // R7LP
                ret_tmp = r7lp_lock(&r7lp, groups[k]);

                break;
            case 18: // R8LP
                ret_tmp = r8lp_lock(&r8lp, groups[k]);

                break;
            case 19: // R9LP
                ret_tmp = r9lp_lock(&r9lp, groups[k]);

                break;
            case 20: // R10LP
                ret_tmp = r10lp_lock(&r10lp, groups[k]);

                break;
            case 21: // R2LP
                ret_tmp = r2lp_lock(&r2lp, groups[k]);

                break;
            }

            lock[i] = ret_tmp[0];
            block[i] = ret_tmp[1];

            free(ret_tmp);
            ret_tmp = NULL;

            /*******code for critical section*******/
            end_time = get_cycles() + (int)(GHZ * cs_lengths[k]);
            while (get_cycles() < end_time)
            {
                cpu_relax();
            }
            /***************************************/
            __sync_synchronize();

            switch (alg)
            {
            case 1: // MCS
                ret_tmp = mcs_unlock(&mcs, &mcs_self);

                break;
            case 2: // RNLP
                ret_tmp = rnlp_unlock(&rnlp, &rnlp_self);

                break;
            case 3: // U-C-RNLP
                ret_tmp = crnlp_uniform_unlock(&crnlp_uniform, ctx->id);

                break;
            case 6: // U-C-RNLP + FGLS
                ret_tmp = crnlp_uniform_fgls_unlock(&crnlp_uniform_fgls, ctx->id);

                break;
            case 7: // U-C-RNLP + FLLS (RR lock)
                ret_tmp = crnlp_uniform_flls_unlock(&crnlp_uniform_flls, ctx->id);

                break;
            case 8: // G-C-RNLP
                ret_tmp = crnlp_general_unlock(&crnlp_general, ctx->id);

                break;
            case 11: // G-C-RNLP + FGLS
                ret_tmp = crnlp_general_fgls_unlock(&crnlp_general_fgls, ctx->id);

                break;
            case 12: // G-C-RNLP + FLLS
                ret_tmp = crnlp_general_flls_unlock(&crnlp_general_flls, ctx->id);

                break;
            case 13: // R3LP
                ret_tmp = r3lp_unlock(&r3lp, groups[k]);

                break;
            case 14: // R4LP
                ret_tmp = r4lp_unlock(&r4lp, groups[k]);

                break;
            case 15: // R5LP
                ret_tmp = r5lp_unlock(&r5lp, groups[k]);

                break;
            case 16: // R6LP
                ret_tmp = r6lp_unlock(&r6lp, groups[k]);

                break;
            case 17: // R7LP
                ret_tmp = r7lp_unlock(&r7lp, groups[k]);

                break;
            case 18: // R8LP
                ret_tmp = r8lp_unlock(&r8lp, groups[k]);

                break;
            case 19: // R9LP
                ret_tmp = r9lp_unlock(&r9lp, groups[k]);

                break;
            case 20: // R10LP
                ret_tmp = r10lp_unlock(&r10lp, groups[k]);

                break;
            case 21: // R2LP
                ret_tmp = r2lp_unlock(&r2lp, groups[k]);

                break;
            }

            unlock[i] = ret_tmp[0];

            free(ret_tmp);
            ret_tmp = NULL;
        }

        ticket_lock(&csvlock);

        for (i = 0; i < BUFSIZE; i++)
        {
            k = l*BUFSIZE + i;

            fprintf(csvfile, "%d, %d, %d, %d, %" PRIu64 ", %" PRIu64 ", %" PRIu64", %lu, %f\n",
                    ctx->id, nr_cpus, nr_resources, nr_requested,
                    lock[i], unlock[i], block[i],
                    requests[k], cs_lengths[k]);

            if(lock[i] > max_l)
                max_l = lock[i];

            if(unlock[i] > max_u)
                max_u = unlock[i];
           
            sum_l += lock[i];
            sum_u += unlock[i];
        }

        ticket_unlock(&csvlock);
    }

    //fprintf(stderr,"Thread %d: max lock %f, max unlock %f\navg lock %f, avg unlock %f\n",
    //        ctx->id, max_l/GHZ, max_u/GHZ, sum_l/(TRIALS*GHZ), sum_u/(TRIALS*GHZ));

    return NULL;
}

int main(int argc, char** argv)
{
    if (argc != 9)
    {
        printf("argc: %d\n",argc);
        fprintf(stderr, "Usage: run cpus nr_requested nr_resources cs_length cs_alpha cs_beta alg csvfile\n");
        return 1;
    }

    int i;
    pthread_t *tasks;
    struct thread_context* ctx;
    
    mlockall(MCL_CURRENT | MCL_FUTURE);
    
    nr_cpus = atoi(argv[1]);
    nr_requested = atoi(argv[2]);
    nr_resources = atoi(argv[3]);
    cs_length = atoi(argv[4]);
    cs_alpha = atoi(argv[5]); // fraction of requests to be shorter, as int in [0,100]
    cs_beta = atoi(argv[6]);  // fraction of L_max for short requests, as int in [0,100]
    alg = atoi(argv[7]);
    csvfile = fopen(argv[8], "w");

    tasks = calloc(nr_cpus, sizeof(pthread_t));
    ctx = malloc(nr_cpus * sizeof(struct thread_context));

    fprintf(csvfile, "thread, numcores, numresources, numrequested, lock, unlock, block, resources\n");

    rnlp_init(&rnlp);
    crnlp_uniform_init(&crnlp_uniform);
    crnlp_uniform_fgls_init(&crnlp_uniform_fgls);
    crnlp_uniform_flls_init(&crnlp_uniform_flls);
    crnlp_general_init(&crnlp_general);
    crnlp_general_fgls_init(&crnlp_general_fgls);
    crnlp_general_flls_init(&crnlp_general_flls);
    r2lp_init(&r2lp);
    r3lp_init(&r3lp);
    r4lp_init(&r4lp);
    r5lp_init(&r5lp);
    r6lp_init(&r6lp);
    r7lp_init(&r7lp);
    r8lp_init(&r8lp);
    r9lp_init(&r9lp);
    r10lp_init(&r10lp);
    ticket_init(&csvlock);
    mcs_init(&mcs);

#ifdef DISABLE_INTERRUPTS
    iopl(3);
#endif

    for(i = 0; i < nr_cpus; i++)
    {
        ctx[i].id = i;
        ctx[i].objects = shared_objects;
    }

    for(i = 0; i < nr_cpus; i++)
    {
        mb();
        pthread_create(&tasks[i], NULL, rt_thread, (void *) (&ctx[i]));
        mb();
    }

    for(i = 0; i < nr_cpus; i++)
    {
        pthread_join(tasks[i], NULL);
    }

    return 0;
}
