#!/usr/bin/env python

from __future__ import division

import argparse
import random
import sys
import math
import time
import traceback
import multiprocessing
import multiprocessing.pool

from collections import defaultdict
from csv import DictWriter
from itertools import product
from math import ceil
from multiprocessing import cpu_count
from numpy import arange
from pprint import pprint
from copy import deepcopy
#from timeout_decorator import timeout

from schedcat.model.tasks import SporadicTask, TaskSystem
from schedcat.generator.tasks import exponential, uniform, multimodal
from schedcat.generator.tasksets import NAMED_PERIODS, NAMED_UTILIZATIONS
from schedcat.util.storage import storage
import schedcat.sched.edf as edf
import schedcat.model.resources as resources
import schedcat.generator.tasks as tasks
import schedcat.locking.bounds as bounds
from schedcat.util.time import ms2us, ms27s, ms28s, ms2ns

from util.stats import BernoulliEstimator

from gurobipy import *
from basic_ilp import *
from multiprocessing import Lock
import time

MIN_SAMPLES = 100
MAX_SAMPLES = 100
MAX_CI      = 0.05
CONFIDENCE  = 0.95
USE_RAND = True
PRINT_OUTPUT = False
TIMING_ONLY = False
TIMING_ALSO = True
BASIC_LIMIT = 0.1
DURATION_LIMIT = 2.7
RW_DURATION_LIMIT = 2.7
OUTFILE = "ilp_out.txt"
convert_func = ms2us
use_overhead = True

#microseconds
if (convert_func == ms2us):
    O_MCS   = 1

    # Overhead values
    O_UCRNLP  = 13
    O_GCRNLP  = 15

    O_CGLP_ARR  = [0, 0, 2, 3, 3, 3, 3, 3, 3, 3, 3]
    # for k of...     1  2  3  4  5  6  7  8  9 10

    NAMED_CS_LENGTHS = { # In micro seconds
        'short'         : uniform(1, 15),
        'moderate'      : uniform(15, 100),
        'xsmall'        : exponential(1,    100,    10),
        'small'         : exponential(10,   1000,   100),
        'large'         : exponential(100,  10000,  1000),
        'long'          : uniform(100, 1000),
        'bimo-medium'   : multimodal([(uniform(15,   100), 5),
                                      (uniform(500, 1000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(15,   100), 7),
                                      (uniform(500, 1000), 3)]),
        'bimo-moderate' : multimodal([(uniform(15,   100), 6),
                                      (uniform(100, 1000), 3)]),
        'bimo-long'     : multimodal([(uniform(15,   100), 4),
                                      (uniform(100, 1000), 5)])
    }
elif (convert_func == ms27s):
# -7-sec
    O_MCS   = 7

    # Overhead values
    O_UCRNLP  = 130
    O_GCRNLP  = 151

    O_CGLP_ARR  = [0, 0, 25, 31, 28, 30, 29, 27, 26, 28, 28]
    # for k of...     1   2   3   4   5   6   7   8   9  10

    NAMED_CS_LENGTHS = { # In -7 seconds
        'short'         : uniform(10, 150),
        'moderate'      : uniform(150, 1000),
        'xsmall'        : exponential(10,    1000,    100),
        'small'         : exponential(100,   10000,   1000),
        'large'         : exponential(1000,  100000,  10000),
        'long'          : uniform(1000, 10000),
        'bimo-medium'   : multimodal([(uniform(150,   1000), 5),
                                      (uniform(5000, 10000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(150,   1000), 7),
                                      (uniform(5000, 10000), 3)]),
        'bimo-moderate' : multimodal([(uniform(150,   1000), 6),
                                      (uniform(1000, 10000), 3)]),
        'bimo-long'     : multimodal([(uniform(150,   1000), 4),
                                      (uniform(1000, 10000), 5)])
    }
elif (convert_func == ms28s):
# e-8 seconds
    O_MCS   = 68

    # Overhead values
    O_UCRNLP  = 1297
    O_GCRNLP  = 1507

    O_CGLP_ARR  = [0, 0, 247, 307, 281, 301, 290, 271, 258, 280, 282]
    # for k of...     1    2    3    4    5    6    7    8    9   10

    NAMED_CS_LENGTHS = { # In nano seconds
        'short'         : uniform(100, 1500),
        'moderate'      : uniform(1500, 10000),
        'xsmall'        : exponential(100,    10000,    1000),
        'small'         : exponential(1000,   100000,   10000),
        'large'         : exponential(10000,  1000000,  100000),
        'long'          : uniform(10000, 100000),
        'bimo-medium'   : multimodal([(uniform(1500,   10000), 5),
                                      (uniform(50000, 100000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(1500,   10000), 7),
                                      (uniform(50000, 100000), 3)]),
        'bimo-moderate' : multimodal([(uniform(1500,   10000), 6),
                                      (uniform(10000, 100000), 3)]),
        'bimo-long'     : multimodal([(uniform(1500,   10000), 4),
                                      (uniform(10000, 100000), 5)])
    }
elif (convert_func == ms2ns):
# nanoseconds
    O_MCS   = 683

    # Overhead values
    O_UCRNLP  = 12966
    O_GCRNLP  = 15067

    O_CGLP_ARR  = [0, 0, 2470, 3066, 2812, 3009, 2897, 2714, 2584, 2802, 2822]
    # for k of...     1     2     3     4     5     6     7     8     9    10

    NAMED_CS_LENGTHS = { # In nano seconds
        'short'         : uniform(1000, 15000),
        'moderate'      : uniform(15000, 100000),
        'xsmall'        : exponential(1000,    100000,    10000),
        'small'         : exponential(10000,   1000000,   100000),
        'large'         : exponential(100000,  10000000,  1000000),
        'long'          : uniform(100000, 1000000),
        'bimo-medium'   : multimodal([(uniform(15000,   100000), 5),
                                      (uniform(500000, 1000000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(15000,   100000), 7),
                                      (uniform(500000, 1000000), 3)]),
        'bimo-moderate' : multimodal([(uniform(15000,   100000), 6),
                                      (uniform(100000, 1000000), 3)]),
        'bimo-long'     : multimodal([(uniform(15000,   100000), 4),
                                      (uniform(100000, 1000000), 5)])
    }

def random_is_read(prob):
    return random.random() < prob

def random_is_read_and_mixed(readProb, mixedProb):
    """
    Returns two values, indicating whether this request is reading
    and whether it is both reading and writing.
    """
    val = random.random()
    if val < readProb:
        return True, False    # reading but not mixed
    elif val < readProb + mixedProb:
        return True, True     # reading and mixed
    else:
        return False, False   # not reading and not mixed

def random_is_nested(prob):
    return random.random() < prob

def get_resource(total):
    return random.randint(0,total-1)

def get_resources(total, needed):
    lst = []
    while len(lst) < needed:
        temp = random.randint(0,total-1)
        if temp not in lst:
            lst.append(temp)
    return lst

def determine_group_count(numRequests, nestedProb, numResources,
                          numResourcesNested, numTasks):
    num_resources_requested = int(numRequests * nestedProb * numResourcesNested + \
                                  numRequests * (1 - nestedProb))
    num_groups_min = max(2, int(math.ceil(num_resources_requested / numResources)))
    return num_groups_min

def create_group_resources(numGroups, numResources, cslenFunc):
    """
    Returns a list of groups, where each group is represented by the
    unclaimed resources for that group.
    """
    resources_by_group = []

    for i in range(numGroups):
        cs_len = int(NAMED_CS_LENGTHS[cslenFunc]())
        resources_by_group.append( (range(numResources), cs_len) )

    return resources_by_group

def get_group_resources(resourcesByGroup, numRequesting, numResources, cslenFunc):
    """
    Randomly chooses a group and numRequesting number of resources from that
    group.  Returns the group number and the list of selected resource numbers.
    """
    group_has_space = []
    for group_num in range(len(resourcesByGroup)):
        if len(resourcesByGroup[group_num][0]) >= numRequesting:
            group_has_space.append(group_num)

    # If there aren't any groups with space, make a new one
    if len(group_has_space) > 0:
        group_num = random.choice(group_has_space)
    else:
        group_num = len(resourcesByGroup)
        group_cs_len = int(NAMED_CS_LENGTHS[cslenFunc]())
        resourcesByGroup.append( (range(numResources), group_cs_len) )

    resources_chosen = []
    for i in range(numRequesting):
        idx = random.randint(0, len(resourcesByGroup[group_num][0])-1)
        resources_chosen.append(resourcesByGroup[group_num][0].pop(idx))

    cs_len = int(random.uniform(0.9, 1.1) * resourcesByGroup[group_num][1])
    #cs_len = resourcesByGroup[group_num][1]

    return group_num, resources_chosen, cs_len

# the following two classes are take from 
# http://stackoverflow.com/questions/6974695/python-process-pool-non-daemonic
# without time, you cannot use timeouts within the threads.
class NoDaemonProcess(multiprocessing.Process):
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

class ExperimentManager(object):
    def __init__(self, parameters, outfile, timingfile=None):
        seed = time.time()
        sys.stderr.write("\nseed: {0}\n".format(seed))
        random.seed(seed)
        self.parameters = parameters
        self.outfile = outfile
        self.out = DictWriter(outfile, parameters.keys()+DesignPoint.metrics)
        self.out.writeheader()
        self.create_design_points()

        if timingfile is not None:
            self.timingfile = timingfile
            self.timingout = DictWriter(timingfile, parameters.keys()+["greedy","basic","duration","rw-duration"]+["numBasicTimeouts","numDurationTimeouts","numRwDurationTimeouts"], delimiter="|")
            self.timingout.writeheader()
        else:
            self.timingfile = None
            self.timingout = None
    
    def valid(self, dp):
        return dp.processors >= dp.sys_util

    def create_design_points(self, shuffle = True):
        self.design_points = filter(self.valid,
            [DesignPoint(**dict(zip(self.parameters.keys(),val)))
                for val in product(*self.parameters.values())])

        if shuffle:
            random.shuffle(self.design_points)

    def run(self, processors, status = True):
        if processors == 1:
            self._run_serial(status)
        else:
            self._run_parallel(processors, status)

        self.finish()

        return # need to return the results here.

    #Useful for debugging purposes.
    def _run_serial(self, status = True):
        rows = map(lambda dp: dp.run(), self.design_points)
        for (i, (row, timingDict)) in enumerate(rows):
            if status and (self.outfile != sys.stdout):
                sys.stderr.write('\rdone {0:%}'.format(i/len(rows)))
            self.out.writerow(row)

    def _run_parallel(self, processors, status = True):
        #pool = Pool(processes = processors)
        pool = MyPool(processes = processors)
        try:
            for (i, (row, timingDict)) in enumerate(pool.imap_unordered(pickleable_function, self.design_points)):
                if status and (self.outfile != sys.stdout):
                    sys.stderr.write('\rdone {0:%}'.format(i/len(self.design_points)))
                self.out.writerow(row)
                if timingDict is not None and self.timingout is not None:
                    self.timingout.writerow(timingDict)
            pool.close()
        except Exception as e:
            pool.terminate()
            print e
            raise

    def finish(self):
        self.outfile.close()
        if self.timingfile:
            self.timingfile.close()

# for whatever reason, imap_unordered will not take a lambda function, because
# it is not pickleable.
def pickleable_function(dp):
    return dp.run()

class DesignPoint(object):
    metrics = ["MCS","RW-RNLP","UCRNLP","GCRNLP", "BasicILP", "DurationILP", "RW-DurationILP", "NOLOCK"]

    # a graph which represents which tests can decide a metric on a given run.
    test_graph = {}

    def __init__(self, **levels):
        self.__dict__.update(levels)
        self.levels = levels
        self.data = dict([(m, BernoulliEstimator()) for m in DesignPoint.metrics])

    def run(self):
        if TIMING_ONLY or TIMING_ALSO:
            self.times = {}
            self.times["greedy"] = []
            self.times["basicIlp"] = []
            self.times["durationIlp"] = []
            self.times["rwDurationIlp"] = []

        while not self.complete():
            if USE_RAND and "CGLP" in self.metrics:
                print "Error: choosing resources randomly does not yield groups for CGLP"
                return -1
            # Generate the task set
            tg = tasks.TaskGenerator(period = NAMED_PERIODS[self.period],
                                        util   = NAMED_UTILIZATIONS[self.task_util])

            ts = tg.make_task_set(max_util = self.sys_util, squeeze = True, time_conversion=convert_func)

            resources.initialize_resource_model(ts)
            bounds.assign_edf_preemption_levels(ts)

            # Make a copy of the task set for each locking protocol being tested
            group_mcs_ts = ts.copy()
            ucrnlp_ts = ts.copy()
            gcrnlp_ts = ts.copy()
            cglp_ts = ts.copy()
            greedy_ts = ts.copy()
            basicilp_ts = ts.copy()
            durationilp_ts = ts.copy()
            rw_durationilp_ts = ts.copy()
            rwrnlp_ts = ts.copy()

            # Initialize the resource groups, using the number of tasks, the
            # percent requesting, and the nested percentage and level to determine
            # how many groups are needed
            num_requests = int(self.req_perc * len(ts))
            num_groups = determine_group_count(num_requests, self.nested_prob,
                                               self.numresources,
                                               self.numresourcesnested,
                                               len(ts))
            resources_by_group = create_group_resources(num_groups,
                                                        self.numresources,
                                                        self.cs_length)

            # Randomly choose self.req_perc percentage of tasks to make requests,
            # and for each of the requesting tasks, determine its request
            # parameters
            all_reqs = []
            for i in range(0, num_requests):
                # Determine whether the request is for a read or write, nested or
                # not, and the CS length
                is_nested = random_is_nested(self.nested_prob)
                if is_nested:
                    is_read, is_mixed = random_is_read_and_mixed(self.read_prob, self.mixed_prob)
                else:
                    # A non-nested request cannot be mixed
                    is_read, is_mixed = random_is_read_and_mixed(self.read_prob, 0)
                if is_nested:
                    num_res = self.numresourcesnested
                else:
                    num_res = 1

                # Choose resources for the request; res_list is a list of resource
                # IDs, each in the range [0, self.numresources)
                if USE_RAND:
                    full_res_list = get_resources(self.numresources, num_res)
                    if is_mixed:
                        # TODO: determine split (for now, even)
                        num_res_read = int(len(full_res_list) / 2)
                        random.shuffle(full_res_list)
                        res_list = [full_res_list[:num_res_read],
                                    full_res_list[num_res_read:]]
                    elif is_read:
                        res_list = [full_res_list, []]
                    else:
                        res_list = [[], full_res_list]
                    cs_len = int(NAMED_CS_LENGTHS[self.cs_length]())
                    all_reqs.append([res_list, cs_len, is_read, is_mixed])
                else:
                    group_num, full_res_list, cs_len = \
                        get_group_resources(resources_by_group, num_res,
                                            self.numresources, self.cs_length)
                    if is_mixed:
                        # TODO: determine split (for now, even)
                        num_res_read = int(len(full_res_list) / 2)
                        random.shuffle(full_res_list)
                        res_list = [full_res_list[:num_res_read],
                                    full_res_list[num_res_read:]]
                    elif is_read:
                        res_list = [full_res_list, []]
                    else:
                        res_list = [[], full_res_list]
                    cs_len = int(NAMED_CS_LENGTHS[self.cs_length]())
                    all_reqs.append([res_list, cs_len, is_read, is_mixed, group_num])

            if PRINT_OUTPUT:
                lock = Lock()
                if lock.acquire(False):
                    outfile = OUTFILE
                    of = open(outfile, "w")
                    for req in all_reqs:
                        of.write(str(req[0]) + ", " + str(req[1]) + ", " + str(req[2]) + \
                                 str(req[3]) + "\n")
                    of.close()
                    lock.release()

            # Test NOLOCK schedulability
            random_tasks = random.sample(range(len(ts)), num_requests)

            temp_reqs = []
            for i in random_tasks:
                req = all_reqs.pop()
                temp_reqs = [req] + temp_reqs
                if USE_RAND:
                    [res_list, cs_len, is_read, is_mixed] = req
                else:
                    [res_list, cs_len, is_read, is_mixed, group_num] = req

                is_just_read = is_read and not is_mixed

                # Update the resource model with the request information
                for res_id in res_list[0] + res_list[1]:
                    ts[i].resmodel[res_id].add_request(cs_len,is_just_read)
            all_reqs = temp_reqs
            for t in ts:
                t.partition = 0
                t.response_time = t.deadline

            confirmed_NOLOCK_schedulable = False
            if not TIMING_ONLY:
                confirmed_NOLOCK_schedulable = edf.is_schedulable(self.processors, ts)
                if not confirmed_NOLOCK_schedulable:
                    for m in DesignPoint.metrics:
                        self.data[m].add_sample(False)
                else:
                    self.data["NOLOCK"].add_sample(True)

            if TIMING_ONLY or confirmed_NOLOCK_schedulable:
                if "Greedy" in self.metrics or "BasicILP" in self.metrics or "DurationILP" in self.metrics or "RW-DurationILP" in self.metrics:
                    greedy_inst = MinColorInstance()
                    greedy_inst.nodes = []
                    for req in all_reqs:
                        if req[2] and not req[3]: # just a read
                            temp_node = Node(req[0][0])
                        else: # mixed or just a write
                            temp_node = Node(req[0][0] + req[0][1]) # treat mixed as writes
                        greedy_inst.nodes.append(temp_node)
                    greedy_inst.n_nodes = len(greedy_inst.nodes)
                    [greedy_time, greedy_groups] = greedy_inst.solve_greedy()
                    num_greedy_groups = max(greedy_groups)
                    if TIMING_ONLY or TIMING_ALSO:
                        self.times["greedy"].append(greedy_time)

                if "BasicILP" in self.metrics:
                    basic_inst = MinColorInstance()
                    basic_inst.nodes = []
                    for req in all_reqs:
                        if req[2] and not req[3]: # just a read
                            temp_node = Node(req[0][0])
                        else: # mixed or just a write
                            temp_node = Node(req[0][0] + req[0][1]) # treat mixed as writes
                        basic_inst.nodes.append(temp_node)
                    basic_inst.n_nodes = len(basic_inst.nodes)
                    [time_setup_basic, time_solve_basic, basic_groups] = solve(basic_inst, num_greedy_groups, BASIC_LIMIT)
                    if basic_groups == -1:
                        print "Error: Turn off debug option in basic_ilp.py"
                    if basic_groups == []:
                        basic_groups = deepcopy(greedy_groups)
                        print "Setting basic_groups"
                    if TIMING_ONLY or TIMING_ALSO:
                        self.times["basicIlp"].append((time_setup_basic, time_solve_basic))

                if "DurationILP" in self.metrics:
                    duration_inst = MinColorInstance(cslenInst=True)
                    duration_inst.nodes = []
                    for req in all_reqs:
                        if req[2] and not req[3]: # just a read
                            temp_node = Node(req[0][0], req[1])
                        else: # mixed or just a write
                            temp_node = Node(req[0][0] + req[0][1], req[1]) # treat mixed as writes
                        duration_inst.nodes.append(temp_node)
                    duration_inst.n_nodes = len(duration_inst.nodes)
                    [time_setup_duration, time_solve_duration, duration_groups] = solve(duration_inst, num_requests, DURATION_LIMIT)
                    if duration_groups == -1:
                        print "Error: Turn off debug option in basic_ilp.py"
                    if duration_groups == []:
                        duration_groups = deepcopy(greedy_groups)
                        print "Setting duration_groups"
                    if TIMING_ONLY or TIMING_ALSO:
                        self.times["durationIlp"].append((time_setup_duration,
                                                          time_solve_duration))

                if "RW-DurationILP" in self.metrics:
                    rw_duration_inst = MinColorInstance(cslenInst=True, rwInst=True)
                    rw_duration_inst.nodes = []
                    for req in all_reqs:
                        if req[2] and not req[3]: # just a read
                            temp_node = Node([], req[1], req[0][0])
                        else: # mixed or just a write
                            temp_node = Node(req[0][1], req[1], req[0][0]) # pass in write-res, cslen, read-res
                        rw_duration_inst.nodes.append(temp_node)
                    rw_duration_inst.n_nodes = len(rw_duration_inst.nodes)
                    [time_setup_rw_duration, time_solve_rw_duration, rw_duration_groups] = solve(rw_duration_inst, num_requests, RW_DURATION_LIMIT)
                    if rw_duration_groups == -1:
                        print "Error: Turn off debug option in basic_ilp.py"
                    if rw_duration_groups == []:
                        rw_duration_groups = deepcopy(greedy_groups)
                    if TIMING_ONLY or TIMING_ALSO:
                        self.times["rwDurationIlp"].append((time_setup_rw_duration,
                                                            time_solve_rw_duration))

                # If we're just doing timings, we only need the ILP times,
                # which we now have, so don't do the schedulability tests
                if TIMING_ONLY:
                    continue # back to the while-not-complete loop

                # Map requesting tasks to the chosen request parameters
                for i in random_tasks:
                    if USE_RAND:
                        if all_reqs == []:
                            print "All_reqs is empty"
                        res_list, cs_len, is_read, is_mixed = all_reqs.pop()
                    else:
                        if all_reqs == []:
                            print "Error: all_reqs is empty"
                        res_list, cs_len, is_read, is_mixed, group_num = all_reqs.pop()

                    if "Greedy" in self.metrics:
                        greedy_group_num = greedy_groups.pop()
                    if "BasicILP" in self.metrics:
                        if basic_groups == []:
                            print "Error: basic_groups is empty"
                        basicilp_group_num = basic_groups.pop()
                    if "DurationILP" in self.metrics:
                        if duration_groups == []:
                            print "Error: duration_groups is empty"
                        durationilp_group_num = duration_groups.pop()
                    if "RW-DurationILP" in self.metrics:
                        if rw_duration_groups == []:
                            print "Error: rw_duration_groups is empty"
                        rw_durationilp_group_num = rw_duration_groups.pop()

                    if use_overhead:
                        ucrnlp_overhead = O_UCRNLP
                        gcrnlp_overhead = O_GCRNLP
                        cglp_overhead   = O_CGLP_ARR[len(resources_by_group)]
                    else:
                        ucrnlp_overhead = 0
                        gcrnlp_overhead = 0
                        cglp_overhead   = 0

                    is_just_read = is_read and not is_mixed

                    # For blocking calculations for the CGLP, the "resource" is
                    # actually the group to which the request belongs
                    if "CGLP" in self.metrics:
                        cglp_ts[i].resmodel[group_num].add_request(cs_len+cglp_overhead,is_just_read)
                    if "Greedy" in self.metrics:
                        greedy_ts[i].resmodel[greedy_group_num].add_request(cs_len+cglp_overhead,is_just_read)
                    if "BasicILP" in self.metrics:
                        basicilp_ts[i].resmodel[basicilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)
                    if "DurationILP" in self.metrics:
                        durationilp_ts[i].resmodel[durationilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)
                    if "RW-DurationILP" in self.metrics:
                        rw_durationilp_ts[i].resmodel[rw_durationilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)

                    # Update the resource model with the request information
                    for res_id in res_list[0] + res_list[1]:
                        j = res_id
                        group_mcs_ts[i].resmodel[res_id].add_request(cs_len+O_MCS,is_just_read)
                        ucrnlp_ts[i].resmodel[res_id].add_request(cs_len+ucrnlp_overhead,is_just_read)
                        gcrnlp_ts[i].resmodel[res_id].add_request(cs_len+gcrnlp_overhead,is_just_read)


                tasksets = {}
                tasksets["MCS"] = group_mcs_ts
                tasksets["UCRNLP"] = ucrnlp_ts
                tasksets["GCRNLP"] = gcrnlp_ts
                tasksets["CGLP"] = cglp_ts
                if "Greedy" in self.metrics:
                   tasksets["Greedy"] = greedy_ts
                if "BasicILP" in self.metrics:
                   tasksets["BasicILP"] = basicilp_ts
                if "DurationILP" in self.metrics:
                   tasksets["DurationILP"] = durationilp_ts
                if "RW-DurationILP" in self.metrics:
                   tasksets["RW-DurationILP"] = rw_durationilp_ts
                tasksets["NOLOCK"] = ts

                for protocol_name in tasksets:
                    if protocol_name != "NOLOCK":
                        protocol_ts = tasksets[protocol_name]
                        for t in protocol_ts:
                            t.partition = 0
                            t.response_time = t.deadline

                # Write some logic to run all your tests here. This is often study
                # specific, as you can exploit domincance results to accelerate
                # things. A classic example, if the task system isn't schedulable
                # without any synchronization, it isn't schedulable with it.
                if not TIMING_ONLY:
                    self.group_mcs(group_mcs_ts)
                    self.ucrnlp(ucrnlp_ts)
                    self.gcrnlp(gcrnlp_ts)
                    if "CGLP" in self.metrics:
                        self.cglp(cglp_ts)
                    if "Greedy" in self.metrics:
                        self.greedy(greedy_ts)
                    if "BasicILP" in self.metrics:
                        self.basicilp(basicilp_ts)
                    if "DurationILP" in self.metrics:
                        self.durationilp(durationilp_ts)
                    if "RW-DurationILP" in self.metrics:
                        self.rw_durationilp(rw_durationilp_ts)
                    self.rw_rnlp(rwrnlp_ts)

        timingSummaryDict = None
        if TIMING_ONLY or TIMING_ALSO:
            #print "\n\nAggregating timing data."
            #print "System util: %0.3f" % self.sys_util
            timingSummaryDict = {}

            greedyTimes = self.times["greedy"]
            basicIlpTimes = [val for val in self.times["basicIlp"] \
                             if val[1] >= 0]
            durationIlpTimes = [val for val in self.times["durationIlp"] \
                                if val[1] >= 0]
            rwDurationIlpTimes = [val for val in self.times["rwDurationIlp"] \
                                  if val[1] >= 0]

            numBasicTimeouts = len(greedyTimes) - len(basicIlpTimes)
            numDurationTimeouts = len(greedyTimes) - len(durationIlpTimes)
            numRwDurationTimeouts = len(greedyTimes) - len(rwDurationIlpTimes)

            timingSummaryDict["numBasicTimeouts"] = numBasicTimeouts
            timingSummaryDict["numDurationTimeouts"] = numDurationTimeouts
            timingSummaryDict["numRwDurationTimeouts"] = numRwDurationTimeouts

            if len(greedyTimes) > 0:
                greedyTimes.sort()
                n = len(greedyTimes)
                p50 = greedyTimes[int(n / 2)]
                p90 = greedyTimes[int(n * 0.9)]
                p95 = greedyTimes[int(n * 0.95)]
                p99 = greedyTimes[int(n * 0.99)]
                timingSummaryDict["greedy"] = (p50, p90, p95, p99)

            for (ilpName, timesList) in zip(["basic", "duration", "rw-duration"],
                                            [basicIlpTimes, durationIlpTimes, rwDurationIlpTimes]):
                if len(timesList) == 0: continue
                #timesList.sort(key = lambda t: t[0] + t[1]) # by total time for now
                timesList.sort(key = lambda t: t[1]) # by solve time
                n = len(timesList)
                p50 = timesList[int(n / 2)]
                p90 = timesList[int(n * 0.9)]
                p95 = timesList[int(n * 0.95)]
                p99 = timesList[int(n * 0.99)]
                timingSummaryDict[ilpName] = (p50, p90, p95, p99)

            #print "Basic timed out", numBasicTimeouts, "times"
            #print "Duration timed out", numDurationTimeouts, "times"
            #print "RW-Duration timed out", numRwDurationTimeouts, "times"

            #for name in ["greedy", "basic", "duration", "rw-duration"]:
            #    print name + ":", timingSummaryDict[name]

            #print

        if (TIMING_ONLY or TIMING_ALSO) and timingSummaryDict:
            timingSummaryDict.update(self.levels)

        return (dict(self.levels.items() + [(k,v.mean) for (k,v) in self.data.iteritems()]),
                timingSummaryDict)

    def group_mcs(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_group_mcs_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["MCS"].add_sample(res)
        return res

    def gcrnlp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_gcrnlp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["GCRNLP"].add_sample(res)
        return res

    def ucrnlp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_ucrnlp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["UCRNLP"].add_sample(res)
        return res

    def rw_durationilp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["RW-DurationILP"].add_sample(res)
        return res

    def durationilp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["DurationILP"].add_sample(res)
        return res

    def basicilp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["BasicILP"].add_sample(res)
        return res

    def greedy(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["Greedy"].add_sample(res)
        return res

    def cglp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["CGLP"].add_sample(res)
        return res

    def rw_rnlp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_rw_rnlp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["RW-RNLP"].add_sample(res)
        return res

    def complete(self):
        # If we're just getting timings, enough of any metric
        # is good enough to stop
        if TIMING_ONLY:
            for ilpName in self.times:
                if len(self.times[ilpName]) >= MIN_SAMPLES:
                    return True
            return False

        for metric, stat in self.data.iteritems():
            if stat.count < MIN_SAMPLES:
                return False
            elif stat.count >= MAX_SAMPLES:
                return True
            elif stat.confidence_interval(CONFIDENCE) > MAX_CI:
                return False
        return True
        
    def create_task_set(self):
        if USE_RAND and "CGLP" in self.metrics:
            print "Error: choosing resources randomly does not yield groups for CGLP"
            return -1
        # Generate the task set
        tg = tasks.TaskGenerator(period = NAMED_PERIODS[self.period],
                                    util   = NAMED_UTILIZATIONS[self.task_util])

        ts = tg.make_task_set(max_util = self.sys_util, squeeze = True, time_conversion=convert_func)

        resources.initialize_resource_model(ts)
        bounds.assign_edf_preemption_levels(ts)

        # Make a copy of the task set for each locking protocol being tested
        group_mcs_ts = ts.copy()
        ucrnlp_ts = ts.copy()
        gcrnlp_ts = ts.copy()
        cglp_ts = ts.copy()
        greedy_ts = ts.copy()
        basicilp_ts = ts.copy()
        durationilp_ts = ts.copy()
        rw_durationilp_ts = ts.copy()
        rwrnlp_ts = ts.copy()

        # Initialize the resource groups, using the number of tasks, the
        # percent requesting, and the nested percentage and level to determine
        # how many groups are needed
        num_requests = int(self.req_perc * len(ts))
        num_groups = determine_group_count(num_requests, self.nested_prob,
                                           self.numresources,
                                           self.numresourcesnested,
                                           len(ts))
        resources_by_group = create_group_resources(num_groups,
                                                    self.numresources,
                                                    self.cs_length)

        # Randomly choose self.req_perc percentage of tasks to make requests,
        # and for each of the requesting tasks, determine its request
        # parameters
        all_reqs = []
        for i in range(0, num_requests):
            # Determine whether the request is for a read or write, nested or
            # not, and the CS length
            is_nested = random_is_nested(self.nested_prob)
            if is_nested:
                is_read, is_mixed = random_is_read_and_mixed(self.read_prob, self.mixed_prob)
            else:
                # A non-nested request cannot be mixed
                is_read, is_mixed = random_is_read_and_mixed(self.read_prob, 0)
            if is_nested:
                num_res = self.numresourcesnested
            else:
                num_res = 1

            # Choose resources for the request; res_list is a list of resource
            # IDs, each in the range [0, self.numresources)
            if USE_RAND:
                full_res_list = get_resources(self.numresources, num_res)
                if is_mixed:
                    # TODO: determine split (for now, even)
                    num_res_read = int(len(full_res_list) / 2)
                    random.shuffle(full_res_list)
                    res_list = [full_res_list[:num_res_read],
                                full_res_list[num_res_read:]]
                elif is_read:
                    res_list = [full_res_list, []]
                else:
                    res_list = [[], full_res_list]
                cs_len = int(NAMED_CS_LENGTHS[self.cs_length]())
                all_reqs.append([res_list, cs_len, is_read, is_mixed])
            else:
                group_num, full_res_list, cs_len = \
                    get_group_resources(resources_by_group, num_res,
                                        self.numresources, self.cs_length)
                if is_mixed:
                    # TODO: determine split (for now, even)
                    num_res_read = int(len(full_res_list) / 2)
                    random.shuffle(full_res_list)
                    res_list = [full_res_list[:num_res_read],
                                full_res_list[num_res_read:]]
                elif is_read:
                    res_list = [full_res_list, []]
                else:
                    res_list = [[], full_res_list]
                cs_len = int(NAMED_CS_LENGTHS[self.cs_length]())
                all_reqs.append([res_list, cs_len, is_read, is_mixed, group_num])

        if PRINT_OUTPUT:
            lock = Lock()
            if lock.acquire(False):
                outfile = OUTFILE
                of = open(outfile, "w")
                for req in all_reqs:
                    of.write(str(req[0]) + ", " + str(req[1]) + ", " + str(req[2]) + \
                             str(req[3]) + "\n")
                of.close()
                lock.release()
        if "BasicILP" in self.metrics or "DurationILP" in self.metrics or "RW-DurationILP" in self.metrics:
            greedy_inst = MinColorInstance()
            greedy_inst.nodes = []
            for req in all_reqs:
                if req[2] and not req[3]: # just a read
                    temp_node = Node(req[0][0])
                else: # mixed or just a write
                    temp_node = Node(req[0][0] + req[0][1]) # treat mixed as writes
                greedy_inst.nodes.append(temp_node)
            greedy_inst.n_nodes = len(greedy_inst.nodes)
            [greedy_time, greedy_groups] = greedy_inst.solve_greedy()
            num_greedy_groups = max(greedy_groups)

        if "BasicILP" in self.metrics:
            basic_inst = MinColorInstance()
            basic_inst.nodes = []
            for req in all_reqs:
                if req[2] and not req[3]: # just a read
                    temp_node = Node(req[0][0])
                else: # mixed or just a write
                    temp_node = Node(req[0][0] + req[0][1]) # treat mixed as writes 
                basic_inst.nodes.append(temp_node)
            basic_inst.n_nodes = len(basic_inst.nodes)
            [time_setup_basic, time_solve_basic, basic_groups] = solve(basic_inst, num_greedy_groups)
            if basic_groups == -1:
                print "Error: Turn off debug option in basic_ilp.py"
            if basic_groups == []:
                basic_groups = deepcopy(greedy_groups)

        if "DurationILP" in self.metrics:
            duration_inst = MinColorInstance(cslenInst=True)
            duration_inst.nodes = []
            for req in all_reqs:
                if req[2] and not req[3]: # just a read
                    temp_node = Node(req[0][0], req[1])
                else: # mixed or just a write
                    temp_node = Node(req[0][0] + req[0][1], req[1]) # treat mixed as writes 
                duration_inst.nodes.append(temp_node)
            duration_inst.n_nodes = len(duration_inst.nodes)
            [time_setup_duration, time_solve_duration, duration_groups] = solve(duration_inst, num_greedy_groups)
            if duration_groups == -1:
                print "Error: Turn off debug option in basic_ilp.py"
            if duration_groups == []:
                duration_groups = deepcopy(greedy_groups)

        if "RW-DurationILP" in self.metrics:
            rw_duration_inst = MinColorInstance(cslenInst=True, rwInst=True)
            rw_duration_inst.nodes = []
            for req in all_reqs:
                if req[2] and not req[3]: # just a read
                    temp_node = Node([], req[1], req[0][0])
                else: # mixed or just a write
                    temp_node = Node(req[0][1], req[1], req[0][0]) # pass in write-res, cslen, read-res
                rw_duration_inst.nodes.append(temp_node)
            rw_duration_inst.n_nodes = len(rw_duration_inst.nodes)
            [time_setup_rw_duration, time_solve_rw_duration, rw_duration_groups] = solve(rw_duration_inst, num_greedy_groups)
            if rw_duration_groups == -1:
                print "Error: Turn off debug option in basic_ilp.py"
            if rw_duration_groups == []:
                rw_duration_groups = deepcopy(greedy_groups)

        # Map requesting tasks to the chosen request parameters
        for i in random.sample(range(len(ts)), num_requests):
            if USE_RAND:
                if all_reqs == []:
                    print "All_reqs is empty"
                res_list, cs_len, is_read, is_mixed = all_reqs.pop()
            else:
                if all_reqs == []:
                    print "Error: all_reqs is empty"
                res_list, cs_len, is_read, is_mixed, group_num = all_reqs.pop()
            if "Greedy" in self.metrics:
                greedy_group_num = greedy_groups.pop()
            if "BasicILP" in self.metrics:
                if basic_groups == []:
                    print "Error: basic_groups is empty"
                basicilp_group_num = basic_groups.pop()
            if "DurationILP" in self.metrics:
                if duration_groups == []:
                    print "Error: duration_groups is empty"
                durationilp_group_num = duration_groups.pop()
            if "RW-DurationILP" in self.metrics:
                if rw_duration_groups == []:
                    print "Error: rw_duration_groups is empty"
                rw_durationilp_group_num = rw_duration_groups.pop()

            if use_overhead:
                ucrnlp_overhead = O_UCRNLP
                gcrnlp_overhead = O_GCRNLP
                cglp_overhead   = O_CGLP_ARR[len(resources_by_group)]
            else:
                ucrnlp_overhead = 0
                gcrnlp_overhead = 0
                cglp_overhead   = 0

            is_just_read = is_read and not is_mixed

            # For blocking calculations for the CGLP, the "resource" is
            # actually the group to which the request belongs
            cglp_ts[i].resmodel[group_num].add_request(cs_len+cglp_overhead,is_just_read)
            if "Greedy" in self.metrics:
                greedy_ts[i].resmodel[greedy_group_num].add_request(cs_len+cglp_overhead,is_just_read)
            if "BasicILP" in self.metrics:
                basicilp_ts[i].resmodel[basicilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)
            if "DurationILP" in self.metrics:
                durationilp_ts[i].resmodel[durationilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)
            if "RW-DurationILP" in self.metrics:
                rw_durationilp_ts[i].resmodel[rw_durationilp_group_num].add_request(cs_len+cglp_overhead,is_just_read)

            # Update the resource model with the request information
            for res_id in res_list[0] + res_list[1]:
                j = res_id
                ts[i].resmodel[res_id].add_request(cs_len,is_just_read)
                group_mcs_ts[i].resmodel[res_id].add_request(cs_len+O_MCS,is_just_read)
                ucrnlp_ts[i].resmodel[res_id].add_request(cs_len+ucrnlp_overhead,is_just_read)
                gcrnlp_ts[i].resmodel[res_id].add_request(cs_len+gcrnlp_overhead,is_just_read)

        tasksets = {}
        tasksets["MCS"] = group_mcs_ts
        tasksets["UCRNLP"] = ucrnlp_ts
        tasksets["GCRNLP"] = gcrnlp_ts
        tasksets["CGLP"] = cglp_ts
        if "Greedy" in self.metrics:
           tasksets["Greedy"] = greedy_ts
        if "BasicILP" in self.metrics:
           tasksets["BasicILP"] = basicilp_ts
        if "DurationILP" in self.metrics:
           tasksets["DurationILP"] = durationilp_ts
        if "RW-DurationILP" in self.metrics:
           tasksets["RW-DurationILP"] = rw_durationilp_ts
        tasksets["NOLOCK"] = ts

        for protocol_name in tasksets:
            protocol_ts = tasksets[protocol_name]
            for t in protocol_ts:
                t.partition = 0
                t.response_time = t.deadline

        return tasksets

def output_task_set(tsi, paramstr, filenum, dp, dpi):
    group_mcs_ts, ucrnlp_ts, gcrnlp_ts, cglp_ts, rwrnlp_ts, org_ts = dp.create_task_set()

    filename = "%s_tasksetid-%d" % (paramstr, tsi)

    f1 = open("tasksets-full/" + ("%03d_" % filenum) + filename + ".txt", 'w')
    f2 = open("tasksets/" + ("%03d" % filenum) + ".txt", 'w')
    f1.write("[DP %d] %s\n" % (dpi, str(dp.levels)))
    for i in range(len(cglp_ts)):
        cslen = max(cglp_ts[i].resmodel.values()[0].max_read_length,
                    cglp_ts[i].resmodel.values()[0].max_write_length)
        f1.write(str(i) + " | " + str(cglp_ts[i].resmodel.keys()[0]) + " | " + str(org_ts[i].resmodel.keys()) + " | " + str(cslen) + "\n")
        f2.write(str(org_ts[i].resmodel.keys()) + "\n")

    f1.close()
    f2.close()

    print "%d tasks" % len(cglp_ts)

# Helper function to more easily specify parameters
def myrange(start, end, inc):
    return arange(start, end+inc, inc)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', "--outfile", type = argparse.FileType('w'),
                        default = sys.stdout,
                        help = "File to output csv file to")
    parser.add_argument('-t', "--timingfile",
                        help = "File to output timing data to")
    parser.add_argument('-p', "--pretend", action='store_true',
                        help = "Only print design point, do not execute")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help="Number of processors to execute on")
    parser.add_argument('-q', "--quiet", action='store_true',
                        help = "Quash all progress and status updates")
    args = parser.parse_args()

    params = storage()

    params.processors = [16]
    params.sys_util = myrange(2, 16, 0.5)
    params.task_util = ['uni-medium']
    params.period = ['uni-short']
    params.cs_length = ['moderate','bimo-medium','bimo-medium2']
    params.req_perc = [1]
    params.read_prob = [0]
    params.mixed_prob = [0]
    params.nested_prob = [0.2]
    params.numgroups = [100]
    params.numresources = [64]
    params.numresourcesnested = [4]

    if TIMING_ONLY or TIMING_ALSO:
        if args.timingfile:
            timingfile = open(args.timingfile, 'w')
        else:
            timingfile = open("timing_out.csv", 'w')
    else:
        timingfile = None

    exp = ExperimentManager(params, args.outfile, timingfile)

    if args.pretend:
        num_task_sets = 50
        filenum = 1
        for dpi in range(len(exp.design_points)):
            dp = exp.design_points[dpi]

            if dp.levels["task_util"] == "uni-heavy":
                dp.levels["cs_length"] = "long"
            elif dp.levels["task_util"] == "uni-medium":
                dp.levels["cs_length"] = "bimo-medium"
            elif dp.levels["task_util"] == "uni-medlight":
                dp.levels["cs_length"] = "xsmall"

            paramstr = "cslen-%s_numcores-%d_taskutil-%s_numgroups-%d_numresources-%d_period-%s_reqperc-%s_sysutil-%s_numrequested-%d_nestedprob-%s_readprob-%s" % \
                (dp.levels["cs_length"],
                 dp.levels["processors"],
                 dp.levels["task_util"],
                 dp.levels["numgroups"],
                 dp.levels["numresources"],
                 dp.levels["period"],
                 dp.levels["req_perc"],
                 dp.levels["sys_util"],
                 dp.levels["numresourcesnested"],
                 dp.levels["nested_prob"],
                 dp.levels["read_prob"],
                 dp.levels["mixed_prob"])

            for tsi in range(num_task_sets):
                output_task_set(tsi, paramstr, filenum, dp, dpi)
                filenum += 1

    if not args.quiet:
        print "Total design points: ", len(exp.design_points)

    if not args.pretend:
        exp.run(args.processors, not args.quiet)

if __name__ == '__main__':
    main()
